/* automatically generated by configure */
/* edit configure.in to change version number */
#define PHP_MAJOR_VERSION 5
#define PHP_MINOR_VERSION 4
#define PHP_RELEASE_VERSION 0
#define PHP_EXTRA_VERSION "RC3-dev"
#define PHP_VERSION "5.4.0RC3-dev"
#define PHP_VERSION_ID 50400
