EXPORTS
php_pcre_compile
php_pcre_copy_substring
php_pcre_exec
php_pcre_get_substring
php_pcre_get_substring_list
php_pcre_info
php_pcre_maketables
php_pcre_study
php_pcre_version
php_pcre_fullinfo
php_pcre_free
